/*
 * TabStopStrategy.m
 * 2/14/99 Mike Trent (don't tell Angela!)
 * mtrent@msn.fullfeed.com
 *
 * This tab-stop code comes from Mike Ferris. I pulled the code from TextExtras
 * some time ago, but I think it's fundamentally the same as the TextEdit
 * example that ships with Mac OS X Server (except not in Java). I wonder why
 * having a uniform mono-spaced-font text view is so hard?
 *
 * I pulled this code into a stragegy object so I can stop cut-pasting this
 * code into all my apps. The strategy is a Singleton class.
 */

#import "TabStopStrategy.h"

static NSArray *tabStopArrayForFontAndTabWidth(NSFont *font, unsigned tabWidth) {
    static NSMutableArray *array = nil;
    static float currentWidthOfTab = -1;
    float charWidth;
    float widthOfTab;
    unsigned i;

    if ([font glyphIsEncoded:(NSGlyph)' ']) {
        charWidth = [font advancementForGlyph:(NSGlyph)' '].width;
    } else {
        charWidth = [font maximumAdvancement].width;
    }
    widthOfTab = (charWidth * tabWidth);

    if (!array) {
        array = [[NSMutableArray allocWithZone:NULL] initWithCapacity:100];
    }

    if (widthOfTab != currentWidthOfTab) {
        [array removeAllObjects];
        for (i = 1; i <= 100; i++) {
            NSTextTab *tab = [[NSTextTab alloc] initWithType:NSLeftTabStopType location:widthOfTab * i];
            [array addObject:tab];
            [tab release];
        }
        currentWidthOfTab = widthOfTab;
    }

    return array;
}

@implementation TabStopStrategy

+ (id)strategy
{
    static TabStopStrategy *sharedInstance = nil;

    if (!sharedInstance) {
        sharedInstance = [[TabStopStrategy alloc] init];
    }

    return sharedInstance;
}

- (void)fixTabStopsInView:(NSTextView*)view
{
    NSRange r = NSMakeRange(0,[[view textStorage] length]);
    [self fixTabStopsInEditedTextStorage:[view textStorage] range:r];
}


- (void)fixTabStopsInEditedTextStorage:(NSTextStorage *)textStorage range:(NSRange)inRange
{
    NSString *string = [textStorage string];

    if ([string length] > 0) {
        // Generally NSTextStorage's attached to plain text NSTextViews only have one font.  But this is not generally true.  To ensure the tabstops are uniform throughout the document we always base them on the font of the first character in the NSTextStorage.
        NSFont *font = [textStorage attribute:NSFontAttributeName atIndex:0 effectiveRange:NULL];

        // Substitute a screen font is the layout manager will do so for display.
        font = [[[textStorage layoutManagers] objectAtIndex:0] substituteFontForFont:font];

        if ([font isFixedPitch]) {
            unsigned tabWidth = 8 ;
            NSArray *desiredTabStops = tabStopArrayForFontAndTabWidth(font, tabWidth);
            NSRange editedRange = inRange;
            NSRange eRange;
            NSParagraphStyle *paraStyle;
            NSMutableParagraphStyle *newStyle;

            editedRange = [string lineRangeForRange:editedRange];

            // We will traverse the edited range by paragraphs fixing the paragraph styles
            while (editedRange.length > 0) {
                paraStyle = [textStorage attribute:NSParagraphStyleAttributeName atIndex:editedRange.location longestEffectiveRange:&eRange inRange:editedRange];
                if (!paraStyle) {
                    paraStyle = [NSParagraphStyle defaultParagraphStyle];
                }
                eRange = NSIntersectionRange(editedRange, eRange);
                if (![[paraStyle tabStops] isEqual:desiredTabStops]) {
                    // Make sure we don't change stuff outside editedRange.
                    newStyle = [paraStyle mutableCopyWithZone:[textStorage zone]];
                    [newStyle setTabStops:desiredTabStops];
                    [textStorage addAttribute:NSParagraphStyleAttributeName value:newStyle range:eRange];
                }
                if (NSMaxRange(eRange) < NSMaxRange(editedRange)) {
                    editedRange.length = NSMaxRange(editedRange) - NSMaxRange(eRange);
                    editedRange.location = NSMaxRange(eRange);
                } else {
                    editedRange = NSMakeRange(NSMaxRange(editedRange), 0);
                }
            }
        }
    }
}


@end
