/*
 * ToolRunner.m
 * 2/9/99 Mike Trent
 * mtrent@msn.fullfeed.com
 *
 * This code comes from a text editor I wrote to help me do my day-job. At Apple, I spend a
 * lot of time maintaining the perl, csh, sh, and makefile scripts that drive our Mac OS X
 * build system(s). I wanted a GUI editor with spif features, such as syntax coloring and
 * autoformatting for perl. In addition, I wanted to run my perl scripts through a syntax
 * checker (perl -c) and to execute all manners of scripts directly in the text editor
 * (vaguely like some kind of MPW shell ... but not really).
 *
 * In my text app, I have some code in _fileActivityNotification that appends incomming text
 * to a scrolling textView. This way, I can see text as it comes in through my script. For
 * the purposes of this app, I just want to get all the goodies at the end. Still, I've
 * included a delegate call in _fileActivityNotification so you can still do this without
 * un-generalizing the code.
 */

#import "ToolRunner.h"

@interface ToolRunner (PrivateTaskStuff)
- (void)_run:(NSString*)filePath args:(NSArray*)args input:(NSString*)inputPath;
- (void)_fileActivityNotification:(NSNotification*)aNotification;
- (void)_taskDeathNotification:(NSNotification*)aNotification;
@end

@interface ToolRunnerDelegate
- (void)toolDidReceiveString:(NSString*)string;
- (void)toolDidTerminate:(int)status;
@end

@implementation ToolRunner

- (void)dealloc
{
    [[NSNotificationCenter defaultCenter] removeObserver:self];

    if (_task) {
        [_task terminate];
        [_task waitUntilExit];
        [_task release];
        _task = nil;
    }

    if (_file) {
        [_file release];
        _file = nil;
    }

    if (_string) {
        [_string release];
        _string = nil;
    }
    
    [super dealloc];
}

- (id)delegate
{
    return _delegate;
}

- (void)setDelegate:(id)delegate
{
    _delegate = delegate;
}

- (void)run:(NSString*)path
{
    [self _run:path args:nil input:nil];
}

- (void)run:(NSString*)path delegate:(id)delegate
{
    [self setDelegate:delegate];
    [self _run:path args:nil input:nil];
}

- (void)run:(NSString*)path arguments:(NSArray*)args
{
    [self _run:path args:args input:nil];
}

- (void)run:(NSString*)path arguments:(NSArray*)args delegate:(id)delegate
{
    [self setDelegate:delegate];
    [self _run:path args:args input:nil];
}

- (BOOL)running
{
    return (_task != nil);
}

- (NSString*)output
{
    return [NSString stringWithString:_string];
}

//
// Privates
//

- (void)_run:(NSString*)filePath args:(NSArray*)args input:(NSString*)inputPath
{
    NSPipe *pipe = [NSPipe pipe];
    NSFileHandle *in;
    NSMutableDictionary *env;

    if ([self running]) return;
    
    _file = [[pipe fileHandleForReading] retain];
    env = [NSMutableDictionary dictionaryWithDictionary:[[NSProcessInfo processInfo] environment]];
    if (_string) [_string release];
    _string = [[NSMutableString string] retain];

    if (inputPath) {
        in = [NSFileHandle fileHandleForReadingAtPath:inputPath];
    } else {
        in = [NSFileHandle fileHandleWithNullDevice];
    }

    _task = [[NSTask alloc] init];
    if (args) [_task setArguments:args];
    [_task setLaunchPath:filePath];
    [_task setStandardInput:in];
    [_task setStandardOutput:pipe];
    [_task setStandardError:pipe];
    [_task setCurrentDirectoryPath:NSHomeDirectory()];
    [_task setEnvironment:env];

    [[NSNotificationCenter defaultCenter]
        addObserver:self
           selector:@selector (_taskDeathNotification:)
               name:NSTaskDidTerminateNotification
             object:_task];
    [[NSNotificationCenter defaultCenter]
        addObserver:self
           selector:@selector (_fileActivityNotification:)
               name:NSFileHandleReadCompletionNotification
             object:_file];

    [_task launch];
    [_file readInBackgroundAndNotify];
}

// an interesting problem with using delegate notifications for task death
// and input receival is occasionally, the task death message comes before
// the input receival notification. This is pretty annoying. We'll work
// around this by not sending the task-death message

- (void)_fileActivityNotification:(NSNotification*)aNotification
{
    NSData* data = [[aNotification userInfo] objectForKey: NSFileHandleNotificationDataItem];

    if ([data length] > 0) {
        NSString* string = [NSString stringWithCString:[data bytes]
                                                length:[data length]];

        if (_delegate && [_delegate respondsToSelector:@selector(toolDidReceiveString:)]) {
            [(ToolRunnerDelegate*)_delegate toolDidReceiveString:string];
        }
        
        [_string appendString:string];
        [_file readInBackgroundAndNotify];
    } else {
        [[NSNotificationCenter defaultCenter]
            removeObserver:self
                      name:NSFileHandleReadCompletionNotification
                    object:_file];
        [_file release];
        _file = nil;
    }

    if (!_task) {
/*#ifdef DEBUG
        NSLog(@"Looks like we got _fileActivityNotification after _taskDeathNotification");
#endif  */      
        // our task already died, so let's send the notification now.
        if (_delegate && [_delegate respondsToSelector:@selector(toolDidTerminate:)]) {
            [(ToolRunnerDelegate*)_delegate toolDidTerminate:_status];
        }
    }
}

- (void)_taskDeathNotification:(NSNotification*)aNotification
{
    _status = [_task terminationStatus];

    [[NSNotificationCenter defaultCenter]
        removeObserver:self
                  name:NSTaskDidTerminateNotification
                object:_task];
    [_task release];
    _task = nil;

    if (_file) {
        // our file handle is still open, so we'll supress the notice here.
/*#ifdef DEBUG
        NSLog(@"Looks like we got _taskDeathNotification before _fileActivityNotification");
#endif       */ 
    } else {
        if (_delegate && [_delegate respondsToSelector:@selector(toolDidTerminate:)]) {
            [(ToolRunnerDelegate*)_delegate toolDidTerminate:_status];
        }
    }
}

@end
